/*
    Quick Log Entry (qle)
    Copyright (C) 2025  Thomas Cort

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>

#include "band.h"
#include "cabrillo.h"
#include "parse.h"
#include "quiz.h"

#include "tmpl_canada_day.h"
#include "tmpl_cq_wpx_cw.h"
#include "tmpl_fcg_fqp.h"
#include "tmpl_qc_qso_party.h"

char *qle_cabrillo_contests[] = {
	"CANADA-DAY",
	"CQ-WPX-CW",
	"FCG-FQP",
	"QC-QSO-PARTY",
	NULL
};

struct qle_cabrillo_template *qle_templates[] = {
	qle_tmpl_canada_day,
	qle_tmpl_cq_wpx_cw,
	qle_tmpl_fcg_fqp,
	qle_tmpl_qc_qso_party,
	NULL
};

#define QLE_CABRILLO_CONTESTS_LEN (sizeof(qle_cabrillo_contests)/qle_cabrillo_contests[0])

void qle_cabrillo_write_tag(FILE *output, char *tag, char *value) {
	fprintf(output, "%s: %s%s", tag, value, QLE_CABRILLO_EOL);
}

int qle_cabrillo_quiz_text(FILE *output, char *tag, char *question, char *pattern) {

	char buffer[64];
	int rc;

	rc = qle_quiz_text(question, pattern, buffer, 64 - 1);
	if (rc == -1) {
		return -1;
	}

	qle_cabrillo_write_tag(output, tag, buffer);
	return 0;
}

int qle_cabrillo_quiz_choice(FILE *output, char *tag, char *question, char *choices[]) {
	int choice;

	choice = qle_quiz_choice(question, choices);
	if (choice == -1) {
		return -1;
	}

	qle_cabrillo_write_tag(output, tag, choices[choice]);
	return choice;
}

int qle_cabrillo_header(FILE *input, FILE *output, FILE *error) {

	int i;
	int rc;
	int choice;

	if (input == stdin) {
		fprintf(error, "format=cabrillo cannot be used when input comes from (stdin)\n");
		return -1;
	} else if (output == stdout) {
		fprintf(error, "format=cabrillo cannot be used when output comes from (stdout)\n");
		return -1;
	}

	qle_cabrillo_write_tag(output, "START-OF-LOG", CABRILLO_VERSION);
	qle_cabrillo_write_tag(output, "CREATED-BY", PACKAGE_STRING);

	choice = qle_cabrillo_quiz_choice(output, "CONTEST", "Which contest?", qle_cabrillo_contests);
	if (choice == -1) {
		return -1;
	}

	for (i = 0; qle_templates[choice][i].type != QLE_CABRILLO_TEMPLATE_DONE; i++) {

		switch (qle_templates[choice][i].type) {
			case QLE_CABRILLO_TEMPLATE_CHOICE:
				rc = qle_cabrillo_quiz_choice(output, qle_templates[choice][i].tag, qle_templates[choice][i].question, qle_templates[choice][i].args.choice.choices);
				break;
			case QLE_CABRILLO_TEMPLATE_TEXT:
				rc = qle_cabrillo_quiz_text(output, qle_templates[choice][i].tag, qle_templates[choice][i].question, qle_templates[choice][i].args.text.pattern);
				break;
			default:
				rc = -1;
				break;
		}

		if (rc == -1) {
			return -1;
		}
	}

	return 0;
}

char *qle_cabrillo_modes[] = {
	"CW",
	"PH",
	"FM",
	"RY",
	"DG"
};
enum qle_cabrillo_mode_t {
	QLE_CABRILLO_MODE_CW = 0,
	QLE_CABRILLO_MODE_PH = 1,
	QLE_CABRILLO_MODE_FM = 2,
	QLE_CABRILLO_MODE_RY = 3,
	QLE_CABRILLO_MODE_DG = 4
};

char *qle_cabrillo_date(void) {
	char *orig;
	static char date[32];

	orig = qle_fields[QLE_FIELD_INDEX_QSO_DATE].value;
	memset(date, '\0', 32);
	if (orig[0] == '\0') {
		date[0] = '\0';
	}

	/* YYYY-MM-DD */
	date[0] = orig[0];
	date[1] = orig[1];
	date[2] = orig[2];
	date[3] = orig[3];
	date[4] = '-';
	date[5] = orig[4];
	date[6] = orig[5];
	date[7] = '-';
	date[8] = orig[6];
	date[9] = orig[7];
	date[10] = '\0';

	return date;
}

char *qle_cabrillo_mode(void) {

	if (strcmp(qle_fields[QLE_FIELD_INDEX_MODE].value, "CW") == 0) {
		return qle_cabrillo_modes[QLE_CABRILLO_MODE_CW];
	} else if (strcmp(qle_fields[QLE_FIELD_INDEX_MODE].value, "SSB") == 0) {
		return qle_cabrillo_modes[QLE_CABRILLO_MODE_PH];
	} else if (strcmp(qle_fields[QLE_FIELD_INDEX_MODE].value, "FM") == 0) {
		return qle_cabrillo_modes[QLE_CABRILLO_MODE_FM];
	} else if (strcmp(qle_fields[QLE_FIELD_INDEX_MODE].value, "RTTY") == 0) {
		return qle_cabrillo_modes[QLE_CABRILLO_MODE_RY];
	} else {
		return qle_cabrillo_modes[QLE_CABRILLO_MODE_DG];
	}
}

struct qle_band_to_freq {
	char *band;
	char *freq;
};

#define QLE_BAND_TO_FREQ_MAP_LEN (25)
struct qle_band_to_freq qle_band_to_freq_map[QLE_BAND_TO_FREQ_MAP_LEN] = {
	{   "160m",  "1800" },
	{    "80m",  "3500" },
	{    "40m",  "7000" },
	{    "20m", "14000" },
	{    "15m", "21000" },
	{    "10m", "28000" },
	{     "6m",    "50" },
	{     "4m",    "70" },
	{     "2m",   "144" },
	{     "2m",   "144" },
	{  "1.25m",   "222" },
	{   "70cm",   "432" },
	{   "33cm",   "902" },
	{   "23cm",  "1.2G" },
	{   "13cm",  "2.3G" },
	{    "9cm",  "3.4G" },
	{    "6cm",  "5.7G" },
	{    "3cm",   "10G" },
	{ "1.25cm",   "24G" },
	{    "6mm",   "47G" },
	{    "4mm",   "75G" },
	{  "2.5mm",  "122G" },
	{    "2mm",  "134G" },
	{    "1mm",  "241G" },
	{  "submm", "LIGHT" }
};

#define QLE_CABRILLO_ACTUAL_FREQ_BANDS_LEN (6)
char *qle_cabrillo_actual_freq_bands[QLE_CABRILLO_ACTUAL_FREQ_BANDS_LEN] = { "160m", "80m", "40m", "20m", "15m", "10m" };

char *qle_cabrillo_freq(void) {
	int i;
	char *band;
	static char freq[32];

	band = qle_fields[QLE_FIELD_INDEX_BAND].value;
	memset(freq, '\0', 32);

	for (i = 0; i < QLE_CABRILLO_ACTUAL_FREQ_BANDS_LEN; i++) {
		if (strcasecmp(qle_cabrillo_actual_freq_bands[i], band) == 0) {
			double f;

			f = strtod(qle_fields[QLE_FIELD_INDEX_FREQ].value, NULL);

			/* MHz to kHz */
			f = f * 1000.0;

			snprintf(freq, 32 - 1, "%d", (int) f);

			return freq;			
		}
	}

	for (i = 0; i < QLE_BAND_TO_FREQ_MAP_LEN; i++) {
		if (strcasecmp(qle_band_to_freq_map[i].band, band) == 0) {
			return qle_band_to_freq_map[i].freq;
		}
	}

	return "";
}

void qle_cabrillo_record(FILE *output) {

	fprintf(
		output,
		"%s: %-5s %s %s %s %-13s %3s %-6s %-13s %3s %-6s %d%s",
		"QSO",
		qle_cabrillo_freq(),
		qle_cabrillo_mode(),
		qle_cabrillo_date(),
		qle_fields[QLE_FIELD_INDEX_TIME_ON].value,
		qle_fields[QLE_FIELD_INDEX_STATION_CALLSIGN].value,
		qle_fields[QLE_FIELD_INDEX_RST_SENT].value,
		qle_fields[QLE_FIELD_INDEX_STX_STRING].value,
		qle_fields[QLE_FIELD_INDEX_CALL].value,
		qle_fields[QLE_FIELD_INDEX_RST_RCVD].value,
		qle_fields[QLE_FIELD_INDEX_SRX_STRING].value,
		0,
		QLE_CABRILLO_EOL
	);
}

void qle_cabrillo_footer(FILE *output) {
	fprintf(output, "%s: %s%s", "END-OF-LOG", "", QLE_CABRILLO_EOL);
}

