/*
    Quick Log Entry (qle)
    Copyright (C) 2025  Thomas Cort

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "args.h"
#include "fmt.h"
#include "parse.h"
#include "str.h"
#include "util.h"

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>

int main(int argc, char *argv[]) {

	int i;
	int ch;
	int header;
	int overwrite;
	int rc;
	char *input_file;
	char *input_file_realpath;
	char *output_file;
	static char prelude[QLE_TOKEN_SIZE];
	FILE *input;
	FILE *output;
	FILE *error;
	char *mypota;
	char *mysota;
	char *mywwff;
	char after[15] /* strlen("YYYYMMDDHHmmss") + '\0' */;
	char before[15] /* strlen("YYYYMMDDHHmmss") + '\0' */;
	time_t input_mtime;
	enum qle_fmt_t fmt;

	static struct option longopts[] = {
		{ "after", required_argument, NULL, 'a' },
		{ "before", required_argument, NULL, 'b' },
		{ "format", required_argument, NULL, 'f' },
		{ "help", no_argument, NULL, 'h' },
		{ "no-header", no_argument, NULL, 'n' },
		{ "overwrite", no_argument, NULL, 'c' },
		{ "pota", required_argument, NULL, 'p' },
		{ "sota", required_argument, NULL, 's' },
		{ "version", no_argument, NULL, 'V' },
		{ "wwff", required_argument, NULL, 'w' }
	};

	input_file = "-";
	output_file = "-";
	input = stdin;
	output = stdout;
	error = stderr;
	header = 1;
	overwrite = 0;
	input_file_realpath = NULL;
	fmt = QLE_FMT_ADIF;
	memset(prelude, '\0', QLE_TOKEN_SIZE);
	qle_strnlcpy(after,  "00000000000000", 15, 0, 14);
	qle_strnlcpy(before, "99999999999999", 15, 0, 14);

	mypota = "MYPOTA@";
	mysota = "MYSOTA@";
	mywwff = "MYWWFF@";

	while ((ch = getopt_long(argc, argv, "a:b:cf:hnp:s:Vw:", longopts, NULL)) != -1) {
		switch (ch) {
			case 'a':
				memcpy(after, optarg, qle_min(strlen(optarg), strlen(after)));
				break;
			case 'b':
				qle_strnlcpy(before, "00000000000000", 15, 0, 14);
				memcpy(before, optarg, qle_min(strlen(optarg), strlen(before)));
				break;
			case 'f':
				if (strcmp(optarg, "compact") == 0) {
					fmt = QLE_FMT_ADIF_COMPACT;
				} else if (strcmp(optarg, "default") == 0) {
					fmt = QLE_FMT_ADIF;
				} else if (strcmp(optarg, "cabrillo") == 0) {
					fmt = QLE_FMT_CABRILLO;
				} else {
					qle_usage();
					return EXIT_FAILURE;
				}
				break;
			case 'h':
				qle_help();
				return EXIT_SUCCESS;
			case 'n':
				header = 0;
				break;
			case 'c':
				overwrite = 1;
				break;
			case 'p':
				qle_strnlcpy(prelude, mypota, QLE_TOKEN_SIZE, 0, strlen(mypota));
				qle_strnlcpy(prelude + strlen(mypota), optarg, QLE_TOKEN_SIZE - strlen(mypota), 0, strlen(optarg));
				break;
			case 's':
				qle_strnlcpy(prelude, mysota, QLE_TOKEN_SIZE, 0, strlen(mysota));
				qle_strnlcpy(prelude + strlen(mysota), optarg, QLE_TOKEN_SIZE - strlen(mysota), 0, strlen(optarg));
				break;
			case 'V':
				qle_version();
				return EXIT_SUCCESS;
			case 'w':
				qle_strnlcpy(prelude, mywwff, QLE_TOKEN_SIZE, 0, strlen(mywwff));
				qle_strnlcpy(prelude + strlen(mywwff), optarg, QLE_TOKEN_SIZE - strlen(mywwff), 0, strlen(optarg));
				break;
			default:
				qle_usage();
				return EXIT_FAILURE;
		}
	}

	argc -= optind;
	argv += optind;

	if (argc == 1) {
		input_file = argv[0];
	} else if (argc == 2) {
		input_file = argv[0];
		output_file = argv[1];
	}

	if (strcmp(input_file, "-") != 0) {
		struct stat st;

		memset(&st, '\0', sizeof(struct stat));

		input_file_realpath = realpath(input_file, NULL);
		/* OK if it fails */

		rc = stat(input_file, &st);
		if (rc == -1) {
			fprintf(stderr, "Cannot Open Input Stream\n");
			return EXIT_FAILURE;
		}
		input_mtime = st.st_mtime;

		input = fopen(input_file, "r");
		if (input == NULL) {
			fprintf(stderr, "Cannot Open Input Stream\n");
			return EXIT_FAILURE;
		}

	} else {
		input_mtime = time(NULL);
	}

	if (strcmp(output_file, "-") != 0) {

		long pos;

		output = fopen(output_file, overwrite == 1 ? "w" : "a");
		if (output == NULL) {
			fprintf(stderr, "Cannot Open Output Stream\n");
			return EXIT_FAILURE;
		}

		/* determine if existing file so we can supress header */
		pos = ftell(output);
		if (pos == -1) {
			fprintf(stderr, "Output Stream Error\n");
			return EXIT_FAILURE;
		} else if (pos != 0) { /* appending? */
			header = 0;
		}

	}

	fmt = (header == 0 && fmt == QLE_FMT_ADIF) ? QLE_FMT_ADIF_NO_HEADER: fmt;

	rc = qle_process(input, output, error, fmt, prelude, input_mtime, input == stdin ? "(stdin)" : (input_file_realpath == NULL ? input_file : input_file_realpath), after, before);

	fclose(input);
	fclose(output);
	fclose(error);

	return rc == 0 ? EXIT_SUCCESS : EXIT_FAILURE;
}
